/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: dispatch.c,v 1.93 2003/03/29 01:58:55 vjt Exp $ */

#include "sux.h"
#include "h.h"
#include "main.h"
#include "numeric.h"
#include "log.h"
#include "match.h"
#include "network.h"
#include "parse.h"
#include "threads.h"
#include "conf.h"
#include "sux-printf.h"
#include "agents.h"
#include "modules.h"

REMOTE_TABLE_INSTANCE(user);
REMOTE_TABLE_INSTANCE(channel);

REMOTE_MEMPOOL_INSTANCE(cmembers);
REMOTE_MEMPOOL_INSTANCE(links);

static void ping_timeout(void);
static gboolean send_ping(void);

static G_INLINE_FUNC gint compile_sjoin_mode(Mode *, gchar *, gint, gchar **);
static G_INLINE_FUNC void compile_mode_mode(Channel *, gchar **);
static G_INLINE_FUNC void compile_umode(User *, gchar *);

static G_INLINE_FUNC void m_message(User *, gint, gchar **);

static void send_nick_burst(void);

static void free_char_slink(gpointer);

static gboolean delete_squitted_users(gchar *, User *, User **);
static gboolean add_squitted_server(GNode *, User **);

static gboolean dispatch_expire(void);

static struct
{
    gshort capabs;
    gchar passwd[PASSWDLEN + 1];

    GSource *ping_tag;
    time_t firsttime;
    gshort flags;
    
} uplink;

GNode *server_tree;
static gint num_links;

/* -----------------------------------------------------------------------------------
 * callback functions
 * -----------------------------------------------------------------------------------
 */
void m_notice(User *u, gint parc, gchar **parv)
{
    m_message(u, parc, parv);
}

void m_private(User *u, gint parc, gchar **parv)
{
    m_message(u, parc, parv);
}

/*
 * m_topic
 * parv[0]: sender pfx
 * parv[1]: channel
 * parv[2]: topic setter
 * parv[3]: ts
 * parv[4]: topic text
 */
void m_topic(User *u, gint parc, gchar **parv)
{
    Channel *c;
    
    g_return_if_fail(parc == 5);
    
    c = _TBL(channel).get(parv[1]);
    g_return_if_fail(c != NULL);

    strcpy(c->topic, parv[4]);
    strcpy(c->topic_nick, parv[2]);
    c->topic_time = atoi(parv[3]);

    message_dispatch(MESSAGE_TOPIC, c, NULL);
}

void m_ping(User *u, gint parc, gchar **parv)
{
    if(u == me.uplink &&
	    G_UNLIKELY(uplink.flags & (FLAGS_SOBSENT|~FLAGS_BURST)))
    {
	uplink.flags &= ~FLAGS_SOBSENT;
	send_out("BURST %d", net_get_sendq());
    }

    send_out(":%s PONG :%s", me.name, parv[1]);
}

/*
 * PASS pwd :TS
 * CAPAB x y z
 * SERVER sname hops info
 * SVINFO ts ts_min 0 :time
 * :sname GNOTICE :Link established
 * BURST
 * NICK ..
 * NICK ..
 * PING
 * TOPIC
 * AWAY
 * PING
 * BURST sendq
 * PING
 */
void m_pong(User *u, gint parc, gchar **parv)
{
    g_return_if_fail(u != NULL);

    if(G_LIKELY(uplink.ping_tag))
    {
	g_source_del(uplink.ping_tag);
	uplink.ping_tag = NULL;
	return;
    }

    if(G_UNLIKELY(u != me.uplink))
    {
	return;
    }

    if(uplink.flags & FLAGS_USERBURST)
    {
	uplink.flags &= ~FLAGS_USERBURST;
	send_out(":%s GNOTICE :%s has processed user/channel burst, "
		"sending topic burst.", me.name, u->name);

	uplink.flags |= FLAGS_SOBSENT;
	
	/* XXX: send shun/topic burst */
	
	send_out("PING :%s", me.name);
    }
    else if(uplink.flags & FLAGS_TOPICBURST)
    {
	uplink.flags &= ~FLAGS_TOPICBURST;
	send_out(":%s GNOTICE :%s has processed topic burst (synched "
		"to network data).", me.name, u->name);
	send_out("PING :%s", me.name);
    }

    return;
}

/*
 * m_nick
 * parv[0] = sender prefix
 * parv[1] = nickname
 * parv[2] = hopcount when new user; TS when nick change
 * parv[3] = TS
 * ---- new user only below ----
 * parv[4] = umode
 * parv[5] = username
 * parv[6] = hostname
 * parv[7] = server
 * parv[8] = serviceid
 * -- If NICKIP
 * parv[9] = IP
 * parv[10] = ircname
 * -- else
 * parv[9] = ircname
 * -- endif
 */
void m_nick(User *u, gint parc, gchar **parv)
{
    if(parc > 3)
    {
	/* new user.
	 */
	u = _TBL(user).get(parv[1]);

	if(u != NULL)
	{
	    /* huh ? someone is using a nick i own ?
	     * kill him !
	     */
	    send_out(rpl_str(ERR_NICKCOLLISION), me.name, u->name, u->name);
	    send_out(":%s KILL %s :Nick Collision", me.name, u->name);
	    if(u->flags & FLAGS_MYAGENT)
	    {
		/* my agent
		 */
		agent_introduce(u);
		agent_rejoin_channels(u);
	    }

	    return;
	}

	u = _TBL(user).alloc(parv[1]);

	u->ts = strtoul(parv[3], NULL, 10);
	compile_umode(u, parv[4]);
	strcpy(u->username, parv[5]);
	strcpy(u->host, parv[6]);
	u->server = _TBL(user).get(parv[7]);
	g_return_if_fail(u->server != NULL);
	u->ip.s_addr = ntohl(strtoul(parv[9], NULL, 10));
	strcpy(u->info, parv[10]);

	message_dispatch(MESSAGE_NEWNICK, u, NULL);
    }
    else
    {
	/* nick change */
	gchar oldnick[NICKLEN + 1];
	g_return_if_fail(_TBL(user).steal(u));
	
	strcpy(oldnick, u->name);
	strcpy(u->name, parv[1]);
	u->ts = strtoul(parv[2], NULL, 10);

	_TBL(user).put(u);

	if(u->id_timeout)
	{
	    g_source_del(u->id_timeout);
	    u->id_timeout = NULL;
	}

	message_dispatch(MESSAGE_NICKCHANGE, u, oldnick);
    }
}

/*
 * m_error
 * parv[0] = sender prefix
 * parv[1] = message
 */
void m_error(User *u, gint parc, gchar **parv)
{
    g_critical_syslog("%s: %s", parv[0], parv[1]);
}

/*
 * m_quit
 * parv[0] = sender prefix
 * parv[1] = comment
 */
void m_quit(User *u, gint parc, gchar **parv)
{
    exit_client(u);
}

/*
 * m_kill 
 * parv[0] = sender prefix 
 * parv[1] = kill victim 
 * parv[2] = kill path
 */
void m_kill(User *u, gint parc, gchar **parv)
{
    User *victim = _TBL(user).get(parv[1]);

    g_return_if_fail(victim != NULL);

    if(victim->flags & FLAGS_MYAGENT)
    {
	/* no need to modify hash tables,
	 * our agents *cannot* be killed !
	 */

	/* but bail out loud if this happens
	 */
	g_warning("Hey !!! %s!%s@%s Killed my agent %s !!! path: %s. Reintroducing ...",
		u->name, u->username, u->host, parv[1], parv[2]);
	
	agent_introduce(victim);
	agent_rejoin_channels(victim);
    }
    else
    {
	/* some other kind of client was killed .. 
	 * update hash table
	 */
	exit_client(victim);
    }
}

/*
 * m_part 
 * parv[0] = sender prefix 
 * parv[1] = channel
 * parv[2] = Optional part reason
 */
void m_part(User *u, gint parc, gchar **parv)
{
    gchar *pfx = parv[0];
    gchar *chname = parv[1];

    Channel *c;
    
    g_return_if_fail(u != NULL);
    g_return_if_fail(pfx != NULL);
    g_return_if_fail(chname != NULL);

    c = _TBL(channel).get(chname);
    g_return_if_fail(c != NULL);

    remove_user_from_channel(c, u);
}

/*
 * m_kick
 * parv[0] = sender prefix
 * parv[1] = channel
 * parv[2] = client to kick
 * parv[3] = kick comment
 */
void m_kick(User *u, gint parc, gchar **parv)
{
    Channel *c;
    User *victim;
    
    g_return_if_fail(u != NULL);

    c = _TBL(channel).get(parv[1]);
    g_return_if_fail(c != NULL);

    victim = _TBL(user).get(parv[2]);
    g_return_if_fail(victim != NULL);
    
    remove_user_from_channel(c, victim);
}

/*
 * m_motd 
 * parv[0] = sender prefix 
 * parv[1] = servername
 */
void m_motd(User *u, gint parc, gchar **parv)
{
    gint i;
    static const gchar *motd[] = 
    {
	PACKAGE_STRING,
	" ",
	"Includes NickServ, ChanServ and MemoServ services.",
	" ",
	"For a listing of available NickServ commands, type: /ns HELP",
	"For a listing of available ChanServ commands, type: /cs HELP",
	"For a listing of available MemoServ commands, type: /ms HELP",
	" ",
	"Project homepage: http://sf.net/projects/suxserv",
	"Author: Barnaba Marcello <" PACKAGE_BUGREPORT ">",
	"Help files based on SirvNET IRC Services ones, (C) 1998-2001 Trevor Klingbeil.",
	"Some snippets of code come from bahamut, (C) 199?-2003 the DALnet coding team.",
	" ",
	"Thanks to: ",
	"         Aleksandr <aleksandr@openssl.it>        - Tester and doc editor",
	"         arlo      <arlo@sniffo.org>             - Designed Sux Services logo",
	"         awgn      <awgn@antifork.org>           - Lots of ideas",
	"         Gastaman  <gastaman@azzurra.org>        - Suggested me how to put FreeBSD support",
	"         tsk       <tsk@users.sf.net>            - Some ideas",
	NULL
    };

    for (i = 0; motd[i] != NULL; i++)
    {
	send_out(rpl_str(RPL_MOTD), me.name, parv[0], motd[i]);
    }

    send_out(rpl_str(RPL_ENDOFMOTD), me.name, parv[0]);
}

/*
 *  m_server 
 *       parv[0] = sender prefix 
 *       parv[1] = servername 
 *       parv[2] = serverinfo/hopcount 
 *       parv[3] = serverinfo
 */
void m_server(User *u, gint parc, gchar **parv)
{
    /* u == server which introduces this server, or == me.uplink
     * if this is the first SERVER command we receive.
     */
    gchar *name = parv[1];

    if(u->ts == 0)
    {
	/* first server .. check for correct name
	 */
	if(u != _TBL(user).get(name))
	{
	    g_critical("No C/N Lines");
	    return;
	}

	/* check for correct password
	 */
	if(mycmp(uplink.passwd, me.pass))
	{
	    g_critical("Access Denied (password mismatch)");
	    return;
	}

	/* check for needed capabilities
	 */
	if(!(uplink.capabs & NEEDED_CAPABS))
	{
	    g_critical("ERROR :Server does not support all the capabs I need");
	    return;
	}

	/* ok, accepted
	 */
	u->ts = NOW;
	u->flags |= FLAGS_SERVER;
	strcpy(u->info, g_strstrip(parv[3]));

	/* insert into the server tree
	 */
	g_node_insert_after(server_tree, NULL, g_node_new((gpointer)u));
	num_links++;

	if(uplink.ping_tag)
	{
	    g_source_del(uplink.ping_tag);
	    uplink.ping_tag = NULL;
	}

	g_timeout_source_add(PING_FREQUENCY, (GSourceFunc) send_ping, NULL);

	uplink.firsttime = NOW;

	send_out(":%s GNOTICE :Link with %s[unknown@0.0.0.0] established, states: TS",
		me.name, u->name);
	send_out("BURST");

	send_nick_burst();

	uplink.flags |= FLAGS_SOBSENT|FLAGS_BURST;
	send_out("PING %s", me.name);
    }
    else
    {
	/* server behind my uplink 
	 */
	User *srv;
	GNode *parent;

	/* sanity check, the server
	 * we are introducing must *NOT*
	 * exist.
	 */
	srv = _TBL(user).get(parv[1]);
	g_return_if_fail(srv == NULL);

	srv = _TBL(user).alloc(parv[1]);
	srv->mode = atoi(parv[2]);
	strcpy(srv->info, parv[3]);
	srv->ts = NOW;
	srv->flags |= FLAGS_SERVER;

	/* put the new server into
	 * the server tree, under
	 * the caller ( uplink server )
	 */
	parent = g_node_find(server_tree, G_IN_ORDER, G_TRAVERSE_ALL, (gpointer)u);
	g_return_if_fail(u != NULL);
	g_node_insert_after(parent, NULL, g_node_new((gpointer)srv));

	num_links++;
    }
}

/*
 * m_info
 * parv[0] = sender prefix 
 * parv[1] = servername
 */
void m_info(User *u, gint parc, gchar **parv)
{
    gint i;
    time_t up;
    extern GSList *modules;
    GSList *johnny /* walker */;
    
    static const gchar *info[] = 
    {
	PACKAGE_STRING,
	"-",
	"Coded by Barnaba Marcello <" PACKAGE_BUGREPORT ">",
	"-",
	"For assistance using Sux Services, /join #sux",
	"Send bug reports to: " PACKAGE_BUGREPORT "",
	"-",
	NULL
    };

    for (i = 0; info[i] != NULL; i++)
    {
	send_out(rpl_str(RPL_INFO), me.name, parv[0], info[i]);
    }

    up = NOW - me.boot;
    send_out(":%s %d %s :Services uptime: %ld days, %ld:%02ld:%02ld",
	    me.name, RPL_INFO, parv[0], up / 86400, (up / 3600) % 24, (up / 60) % 60, up % 60);
    send_out(":%s %d %s :Birth date: " __DATE__ " at " __TIME__,
	    me.name, RPL_INFO, parv[0]);
    send_out(":%s %d %s :-",
	    me.name, RPL_INFO, parv[0]);

    /* XXX: oper-only section !
     */ 
    send_out(":%s %d %s :SQL: Connected to %s [TID: %ld]",
	    me.name, RPL_INFO, parv[0], sql_get_host_info(), sql_thread_id());
    send_out(":%s %d %s :SQL: Protocol v.%d on server version %s",
	    me.name, RPL_INFO, parv[0], sql_get_proto_info(), sql_get_server_info());
    send_out(":%s %d %s :-",
	    me.name, RPL_INFO, parv[0]);
    send_out(":%s %d %s :%d modules loaded:",
	    me.name, RPL_INFO, parv[0], g_slist_length(modules));
    for(johnny = modules; johnny; johnny = g_slist_next(johnny))
    {
	ModDesc *desc = (ModDesc *)johnny->data;

	send_out(":%s %d %s :%s, version: %s [%s]",
		me.name, RPL_INFO, parv[0], desc->name, desc->version, desc->info);
    }

    send_out(rpl_str(RPL_ENDOFINFO), me.name, parv[0]);
}

/*
 * 
 * m_stats 
 * parv[0] = sender prefix 
 * parv[1] = statistics selector
 * parv[2] = server name (current server defaulted, if omitted) 
 */
void m_stats(User *u, gint parc, gchar **parv)
{
    gchar stat = parc > 1 ? parv[1][0] : '*';
    gchar *up_name, *up_port;

    up_name = find_conf_id(NULL, SYMBOL_UPLINK_NAME);
    up_port = find_conf_id(NULL, SYMBOL_UPLINK_PORT);
    
    switch(stat)
    {
	case 'C':
	case 'c':
	    send_out(rpl_str(RPL_STATSCLINE), me.name, parv[0],
		    up_name, up_port);
	    send_out(rpl_str(RPL_STATSNLINE), me.name, parv[0],
		    up_name, "0");
	    break;

	case 'h':
	case 'H':
	    send_out(rpl_str(RPL_STATSHLINE), me.name, parv[0],
		    'H', "*", up_name, 0, -1);
	    break;

	case 'U':
	    send_out(rpl_str(RPL_STATSULINE), me.name, parv[0],
		    'U', me.name, "*", 0, -1);
	    break;

	case 'u':
	    {
		time_t t;
		
		t = NOW - me.boot;

		send_out(rpl_str(RPL_STATSUPTIME), me.name, parv[0],
			t / 86400, (t / 3600) % 24, (t / 60) % 60, t % 60);
	    }
	    break;

	default:
	    stat = '*';
	    break;
    }
    send_out(rpl_str(RPL_ENDOFSTATS), me.name, parv[0], stat);
}

void m_version(User *u, gint parc, gchar **parv)
{
    send_out(rpl_str(RPL_VERSION),
	    me.name, parv[0], PACKAGE_STRING);
}

/*
 * m_squit
 * parv[0] = sender prefix 
 * parv[1] = server name 
 * parv[2] = comment
 */
void m_squit(User *u, gint parc, gchar **parv)
{
    User *srv;
    GNode *n;
    User **squitted_servers = g_new0(User *, num_links + 1);
    gint i;

    /* we don`t care if our uplink squitted us,
     * `cause if this is the case we will receive an
     * ERROR:, and even if this is not the case, it
     * will close our connection
     */
    g_return_if_fail(u != NULL);
    
    srv = _TBL(user).get(parv[1]);
    g_return_if_fail(srv != NULL);

    /* quit all the clients on this link, and on the
     * servers behind it.
     */

    /* n is the root node for this server
     */
    n = g_node_find(server_tree, G_IN_ORDER, G_TRAVERSE_ALL, (gpointer) srv);
    g_node_traverse(n, G_IN_ORDER, G_TRAVERSE_ALL, -1,
	    (GNodeTraverseFunc) add_squitted_server, (gpointer) squitted_servers);
    g_node_destroy(n);

    g_hash_table_foreach_remove(_TBL(user).__hash_tbl, 
	    (GHRFunc) delete_squitted_users, squitted_servers);

    for(i = 0; (srv = squitted_servers[i]); i++)
    {
	_TBL(user).del(srv);
    }

    g_free((gpointer)squitted_servers);
}

/*
 * m_pass 
 * parv[0] = sender prefix 
 * parv[1] = password
 * parv[2] = optional extra version information
 */
void m_pass(User *u, gint parc, gchar **parv)
{
    /* only our uplink can send us a PASS command
     */
    g_return_if_fail(u == me.uplink);

    g_strlcpy(uplink.passwd, parv[1], sizeof(uplink.passwd));
}

/*
 * m_svinfo 
 *       parv[0] = sender prefix 
 *       parv[1] = TS_CURRENT for the server 
 *       parv[2] = TS_MIN for the server 
 *       parv[3] = server is standalone or connected to non-TS only 
 *       parv[4] = server's idea of UTC time
 */
void m_svinfo(User *u, gint parc, gchar **parv)
{
    time_t deltat, theirtime;
    guint their_ts_ver, their_ts_min_ver;

    g_return_if_fail(parc > 4);

    theirtime = atol(parv[4]);

    deltat = ABS(theirtime - NOW);

    if(deltat > 45)
    {
	/* delta too high, drop the link
	 */
	g_critical("Link %s dropped, excessive TS delta (%ld)", u->name, deltat);
	return;
    }
    else if(deltat > 15)
    {
	/* delta high, bail out
	 */
	g_warning("Link %s notable TS delta (my TS=%ld, their TS=%ld, delta=%ld",
		u->name, NOW, theirtime, deltat);
    }

    their_ts_ver = atoi(parv[1]);
    their_ts_min_ver = atoi(parv[2]);
    if(their_ts_ver < SUX_MIN_TS || SUX_CUR_TS < their_ts_min_ver)
    {
	/* update your server =P
	 */
	g_critical_syslog("Incompatible TS version (%d,%d)",
		their_ts_ver, their_ts_min_ver);
	return;
    }

    me.uplink->ts = theirtime;

    return;
}

/*
 * m_join 
 * parv[0] = sender prefix
 * parv[1] = channel
 * parv[2] = channel password (key)
 */
void m_join(User *u, gint parc, gchar **parv)
{
    /* handle only `:user JOIN 0' case
     */
    g_return_if_fail(*parv[1] == '0');

    g_return_if_fail(u->channels != NULL);

    while(u->channels)
    {
	remove_user_from_channel(((SLink*)u->channels->data)->v.c, u);
    }

    return;
}

/* m_mode
 * parv[0] - sender
 * parv[1] - channel
 * parv[2] - TS if chmode, mode string if umode
 * parv[3] - mode string if chmode, NULL if umode
 * parv[4+n] - mode args if chmode, invalid if umode
 */
void m_mode(User *u, gint parc, gchar **parv)
{
    Channel *c;
    time_t tsmode;
    
    g_return_if_fail(parc > 2);

    /* if this is an user mode ...
     */
    if(*parv[1] != '#')
    {
	compile_umode(u, parv[1]);
	return;
    }

    /* else this is a channel mode
     */
    c = _TBL(channel).get(parv[1]); 
    g_return_if_fail(c != NULL);

    /* do tsmode checks
     */
    tsmode = atol(parv[2]);
    if(tsmode != 0 && (tsmode > c->ts))
    {
	g_warning("Ignoring MODE for %s (oldts: %ld, newts: %ld)",
		parv[1], c->ts, tsmode);
	return;
    }

    compile_mode_mode(c, parv);
}

/*
 * m_sjoin 
 * parv[0] - sender 
 * parv[1] - TS 
 * parv[2] - channel 
 * parv[3] - modes + n arguments (key and/or limit) 
 * parv[4+n] - flags+nick list (all in one parameter)
 */
void m_sjoin(User *u, gint parc, gchar **parv)
{
    gint ts;
    gchar *channel;
    gchar *user;
    gchar *users = NULL;
    gchar *modes;
    Channel *c = NULL;
    
    gint i, count;
    gchar *users_arr[512];

    g_return_if_fail(parv != NULL);

    ts = atoi(parv[1]);
    channel = parv[2];

    if(parc == 3)
    {
	/* client sjoin with TS
	 */
	User *u;

	user = parv[0];

	/* we *must* have this one
	 */
	c = _TBL(channel).get(channel);
	g_return_if_fail(c != NULL);
	
	/* and this one, too
	 */
	u = _TBL(user).get(user);
	g_return_if_fail(u != NULL);

	c->ts = ts;
	add_user_to_channel(u, c, 0);

	return;
    }
    else
    {
	gboolean created = FALSE;

	/* server sjoin with modes and users
	 */
	modes = parv[3];

	c = _TBL(channel).get(channel);

	if(*modes != '0')
	{
	    /* server sent channel modes
	     */
	    gint args = 0;

	    if(c == NULL)
	    {
		/* i don`t have this !
		 */
		c = _TBL(channel).alloc(channel);

		c->ts = ts;
		c->bans = NULL;
		c->members = NULL;

		created = TRUE;
	    }

	    /* always keep their modes
	     */
	    if((args = compile_sjoin_mode(&c->mode, modes, parc, parv)) < 0)
	    {
		g_warning("Failed mode compilation on `%s', channel desynched !", modes);
		memset(&c->mode, 0x0, sizeof(Mode));
	    }
	    users = parv[4 + args];

	    if(created)
	    {
		message_dispatch(MESSAGE_NEWCHAN, c, NULL);
	    }
	}
	else
	{
	    users = parv[4];
	}

    }

    /* we *must* have some users
     */
    g_return_if_fail(users != NULL);

    /* so, split the last arg and add
     * all of them ..
     */
    my_g_strsplit(users, ' ', 512, &count, users_arr);
    for(i = 0; i < count; i++)
    {
	gint fl = 0;
	
	user = users_arr[i];

	if(*user == '\0')
	    continue;

	/* .. taking care of their flags
	 */
	if(user[0] == '@' || user[1] == '@')
	    fl |= MODE_CHANOP;

	if(user[0] == '+' || user[1] == '+')
	    fl |= MODE_VOICE;

	while(*user == '@' || *user == '+')
	    user++;
	
	add_user_to_channel(_TBL(user).get(user), c, fl);
    }
}

void m_capab(User *u, gint parc, gchar **parv)
{
    gint i;

    /* okay, is dirty, but this command is issued
     * only *ONCE* in our entire life .. an hash
     * table looked to me so exaggerated ...
     */
    for(i = 1; i < parc; i++)
    {
	if (!mycmp(parv[i], "NOQUIT"))
	    uplink.capabs |= CAPAB_NOQUIT;
	else if (!mycmp(parv[i], "BURST"))
	    uplink.capabs |= CAPAB_BURST;
	else if (!mycmp(parv[i], "UNCONNECT"))
	    uplink.capabs |= CAPAB_UNCONN;
	else if (!mycmp(parv[i], "DKEY"))
	    uplink.capabs |= CAPAB_DKEY;
	else if (!mycmp(parv[i], "ZIP"))
	    uplink.capabs |= CAPAB_ZIP;
	else if (!mycmp(parv[i], "NICKIP"))
	    uplink.capabs |= CAPAB_NICKIP;
	else if (!mycmp(parv[i], "TSMODE"))
	    uplink.capabs |= CAPAB_TSMODE;
    }
}

void m_lusers(User *u, gint parc, gchar **parv)
{
    g_return_if_fail(u != NULL);
    
    /* count channels ...
     */
    send_out(rpl_str(RPL_LUSERCHANNELS),
	    me.name, u->name, _TBL(channel).count());
    
    /* and users ...
     */
    send_out(rpl_str(RPL_GLOBALUSERS),
	    me.name, u->name, _TBL(user).count());

    return;
}

/*
 * m_whois 
 * parv[0] = sender prefix 
 * parv[1] = my name
 * parv[2] = nickname
 */
void m_whois(User *u, gint parc, gchar **parv)
{
    User *target;
    gchar *nick = parv[2];
    gchar chanbuf[BUFSIZE];
    gint len, mlen;
    GSList *johnny; /* walker */
    Channel *c;
    SLink *cm;

    g_return_if_fail(u != NULL);

    target = _TBL(user).get(nick);

    if(target == NULL)
    {
	send_out(rpl_str(ERR_NOSUCHNICK),
		me.name, parv[0], nick);

	return;
    }

    /* send out first line ..
     */
    send_out(rpl_str(RPL_WHOISUSER), me.name,
	    parv[0], target->name, target->username,
	    target->host, target->info);

    /* .. and then channels line(s) ..
     */
    mlen = strlen(me.name) + strlen(parv[0]) + 6 + strlen(nick);
    for(len = 0, *chanbuf = '\0', johnny = target->channels;
	    johnny;
	    johnny = g_slist_next(johnny))
    {
	cm = (SLink*)johnny->data;
	c = cm->v.c;

	if(len + strlen(c->chname) > (gsize) BUFSIZE - 4 - mlen)
	{
	    send_out(":%s %d %s %s :%s",
		    me.name, RPL_WHOISCHANNELS,
		    parv[0], nick, chanbuf);
	    *chanbuf = '\0';
	    len = 0;
	}
	if(cm->flags & MODE_CHANOP)
	    chanbuf[len++] = '@';
	else if(cm->flags & MODE_VOICE)
	    chanbuf[len++] = '+';

	strcpy(chanbuf + len, c->chname);
	len += strlen(c->chname);
	chanbuf[len++] = ' ';
	chanbuf[len] = '\0';
    }
    if(chanbuf[0] != '\0')
    {
	send_out(rpl_str(RPL_WHOISCHANNELS),
		me.name, parv[0], nick, chanbuf);
    }

    /* .. and then server line .. 
     */
    send_out(rpl_str(RPL_WHOISSERVER),
	    me.name, parv[0], nick, target->server->name, target->server->info);

    /* .. and then, all done !
     */
    send_out(rpl_str(RPL_ENDOFWHOIS),
	    me.name, parv[0], nick);
}

/* 
 * m_burst
 *      parv[0] = sender prefix
 *      parv[1] = SendQ if an EOB
 */
void m_burst(User *u, gint parc, gchar **parv)
{
    g_return_if_fail(u != NULL);

    if(parc == 2)
    {
	/* this is an EOB
	 */
	time_t synch_time;
	
	uplink.flags &= ~FLAGS_EOBRECV;
	if(uplink.flags & (FLAGS_SOBSENT|FLAGS_BURST))
	    return;
	
	synch_time = NOW - uplink.firsttime;

	send_out(":%s GNOTICE :synch to %s in %ld sec%s at %s sendq",
		me.name, u->name, synch_time, synch_time == 1 ? "" : "s", parv[1]);

	sql_unlock_tables();
    }
    else
    {
	/* this is a SOB 
	 */
	uplink.flags |= FLAGS_EOBRECV;

	sql_lock_tables();
    }
}

/*
 * m_away 
 * parv[0] = sender prefix 
 * parv[1] = away message
 */
void m_away(User *u, gint parc, gchar **parv)
{
    g_return_if_fail(u != NULL);
    
    /* we don`t care about messages, only about
     * is/is not away
     */
    if(parc < 2 || !*parv[1])
    {
	u->flags &= ~FLAGS_AWAY;
    }
    else
    {
	u->flags |= FLAGS_AWAY;
    }
}

void m_time(User *u, gint parc, gchar **parv)
{
    gchar timebuf[26];
    
    /* screw the original server time format,
     * i think that a ctime_r()is more than enough
     */
    sux_ctime_r(&NOW, timebuf, sizeof(timebuf));
    timebuf[24] = '\0';
    
    send_out(rpl_str(RPL_TIME), me.name, parv[0], me.name, timebuf);
}

void m_admin(User *u, gint parc, gchar **parv)
{
    /* these will be changed in some point ...
     */
    send_out(rpl_str(RPL_ADMINME), me.name, parv[0], me.name);
    send_out(rpl_str(RPL_ADMINLOC1), me.name, parv[0], PACKAGE_NAME);
    send_out(rpl_str(RPL_ADMINLOC2), me.name, parv[0], "Release " PACKAGE_VERSION);
    send_out(rpl_str(RPL_ADMINEMAIL), me.name, parv[0], "coded by " PACKAGE_BUGREPORT);
}

/* convenient agent messages (only if SERVICESHUB is defined in uplink ircd)
 */
void m_cs(User *u, gint parc, gchar **parv)
{
    message_dispatch(MESSAGE_CSREQ, u, parv[1]);
}

void m_ns(User *u, gint parc, gchar **parv)
{
    message_dispatch(MESSAGE_NSREQ, u, parv[1]);
}

void m_ms(User *u, gint parc, gchar **parv)
{
    message_dispatch(MESSAGE_MSREQ, u, parv[1]);
}

void m_os(User *u, gint parc, gchar **parv)
{
    message_dispatch(MESSAGE_OSREQ, u, parv[1]);
}

void m_rs(User *u, gint parc, gchar **parv)
{
    message_dispatch(MESSAGE_RSREQ, u, parv[1]);
}

/* -----------------------------------------------------------------------------------
 * public functions
 * -----------------------------------------------------------------------------------
 */

/* channel-related functions
 */
G_INLINE_FUNC SLink *find_userchannel(User *u, Channel *c)
{
    register GSList *johnny = u->channels; /* walker */

    /* find a channel into an user chanlist
     */
    while(johnny)
    {
	if(((SLink*)johnny->data)->v.c == c)
	    return ((SLink*)johnny->data);

	johnny = g_slist_next(johnny);
    }
    return NULL;
}

G_INLINE_FUNC ChanMember *find_chanmember(Channel *c, User *u)
{
    register GSList *johnny = c->members; /* walker */

    /* find an user into a channel memberlist
     */
    while(johnny)
    {
	if(((ChanMember*)johnny->data)->u == u)
	    return ((ChanMember*)johnny->data);
	johnny = g_slist_next(johnny);
    }

    return NULL;
}

gboolean add_user_to_channel(User *u, Channel *c, guint flags)
{
    ChanMember *cm;
    SLink *lp;

    g_return_val_if_fail(find_chanmember(c, u) == NULL, FALSE);
    g_return_val_if_fail(find_userchannel(u, c) == NULL, FALSE);

    cm = g_mem_chunk_alloc0(_MPL(cmembers));
    lp = g_mem_chunk_alloc0(_MPL(links));

    g_return_val_if_fail(c != NULL, FALSE);
    g_return_val_if_fail(u != NULL, FALSE);
    g_return_val_if_fail(cm != NULL, FALSE);

    cm->u = u;
    cm->flags = flags;

    lp->v.c = c;
    lp->flags = flags;

    u->channels = g_slist_prepend(u->channels, lp);
    c->members = g_slist_prepend(c->members, cm);

    if(G_LIKELY(!(u->flags & FLAGS_MYAGENT)))
    {
	/* our agents do not need channel access,
	 * they have super cow powers !
	 */
	message_dispatch(MESSAGE_JOIN, c, cm);
    }

    return TRUE;
}

gboolean remove_user_from_channel(Channel *c, User *u)
{
    ChanMember *cm = find_chanmember(c, u);
    SLink *lp = find_userchannel(u, c);

    g_return_val_if_fail(cm != NULL, FALSE);
    g_return_val_if_fail(lp != NULL, FALSE);
    
    c->members = g_slist_remove(c->members, cm);
    g_mem_chunk_free(_MPL(cmembers), cm);

    u->channels = g_slist_remove(u->channels, lp);
    g_mem_chunk_free(_MPL(links), lp);

    if(c->members == NULL)
    {
	_TBL(channel).del(c);
    }

    return TRUE;
}

/* start server<->server negotiation
 */
void irc_nego_start(void)
{
    gchar *up_name;

    send_out("PASS %s :TS", me.pass);
    send_out("CAPAB BURST NOQUIT SSJOIN UNCONNECT NICKIP TSMODE");
    send_out("SVINFO %d %d 0 :%lu", SUX_CUR_TS, SUX_MIN_TS, NOW);
    send_out("SERVER %s 1 :%s", me.name, me.info);

    uplink.ping_tag = g_timeout_source_add(PING_TIMEOUT * 2, (GSourceFunc) ping_timeout, NULL);
    
    up_name = find_conf_id(NULL, SYMBOL_UPLINK_NAME);
    if(!up_name)
    {
	g_critical("Uplink name not present in config file");
	return;
    }

    /* placeholder for my uplink
     */
    me.uplink = _TBL(user).alloc(up_name);
    me.uplink->ts = 0;

    me.table_ptr = _TBL(user).alloc(me.name);
    strcpy(me.table_ptr->info, me.info);
    me.table_ptr->ts = NOW;
    me.table_ptr->mode = 0;
    me.table_ptr->flags |= FLAGS_SERVER;

    server_tree = g_node_new((gpointer) me.table_ptr);
    num_links = 1;

    g_timeout_source_add(60 gs_minutes, (GSourceFunc) dispatch_expire, NULL);
}

void exit_client(User *u)
{
    register SLink *lp;

    g_return_if_fail(u != NULL);

    while(u->channels)
    {
	lp = (SLink *)u->channels->data;
	g_return_if_fail(lp->v.c != NULL);

	remove_user_from_channel(lp->v.c, u);
    }

    if(u->id_chans)
    {
	g_slist_foreach(u->id_chans, (GFunc) free_char_slink, NULL);
	g_slist_free(u->id_chans);
    }

    if(u->id_nicks)
    {
	g_slist_foreach(u->id_nicks, (GFunc) free_char_slink, NULL);
	g_slist_free(u->id_nicks);
    }

    if(u->id_timeout)
    {
	g_source_del(u->id_timeout);
	u->id_timeout = NULL;
    }

    _TBL(user).del(u);
}

/* ------------------------------------------------------------------------------------------
 * private functions
 * ------------------------------------------------------------------------------------------
 */

/* ping functions
 */
static void ping_timeout(void)
{
    /* ping timeout
     */
    g_critical_syslog("Ping Timeout");
}

static gboolean send_ping(void)
{
    if(uplink.ping_tag)
    {
	ping_timeout();

	return FALSE;
    }

    send_out("PING :%s", me.name);

    uplink.ping_tag = g_timeout_source_add(PING_TIMEOUT, (GSourceFunc) ping_timeout, NULL);

    return TRUE;
}

/* umode functions
 */

/*
 * half stolen from bahamut/src/channel.c
 */
static G_INLINE_FUNC gint compile_sjoin_mode(Mode *m, gchar *s, gint parc, gchar **parv)
{
    gint args = 0;

    memset(m, 0x0, sizeof(Mode));
    
    /* parse all the mode string ...
     */
    while(G_LIKELY(*s))
    {
	switch(*(s++))
	{
	    case 'i':
		m->mode |= MODE_INVITEONLY;
		break;
	    case 'n':
		m->mode |= MODE_NOPRIVMSGS;
		break;
	    case 'p':
		m->mode |= MODE_PRIVATE;
		break;
	    case 's':
		m->mode |= MODE_SECRET;
		break;
	    case 'm':
		m->mode |= MODE_MODERATED;
		break;
	    case 't':
		m->mode |= MODE_TOPICLIMIT;
		break;
	    case 'r':
		m->mode |= MODE_REGISTERED;
		break;
	    case 'R':
		m->mode |= MODE_REGONLY;
		break;
	    case 'M':
		m->mode |= MODE_MODREG;
		break;
	    case 'c':
		m->mode |= MODE_NOCOLOR;
		break;
	    case 'd':
		m->mode |= MODE_NONICKCHG;
		break;
	    case 'O':
		m->mode |= MODE_OPERONLY;
		break;
	    case 'k':
		m->mode |= MODE_KEY;
		g_strlcpy(m->key, parv[4 + args], KEYLEN + 1);
		args++;
		if (parc < 5 + args)
		    return -1;
		break;
	    case 'l':
		m->mode |= MODE_LIMIT;
		m->limit = atoi(parv[4 + args]);
		args++;
		if (parc < 5 + args)
		    return -1;
		break;
	}
    }

    return args;
}

/*
 * half stolen from bahamut/src/channel.c
 */
static G_INLINE_FUNC void compile_mode_mode(Channel *c, gchar **parv)
{
    User *u;
    ChanMember *cm;
    SLink *lp;

    Mode *m = &c->mode;
    gchar change = '+';
    gchar *s = parv[3];
    gint args = 4;
    gint i;

    static const gint flags[] = 
    {
	MODE_PRIVATE, 'p', MODE_SECRET, 's',
	MODE_MODERATED, 'm', MODE_NOPRIVMSGS, 'n',
	MODE_TOPICLIMIT, 't', MODE_REGONLY, 'R',
	MODE_INVITEONLY, 'i', MODE_NOCOLOR, 'c', MODE_OPERONLY, 'O',
	MODE_MODREG, 'M', MODE_NONICKCHG, 'd',
	0x0, 0x0
    };

    while(G_LIKELY(*s))
    {
	switch(*s)
	{
	    case '+':
	    case '-':
		change = *s;
		break;
	    case 'o':
	    case 'v':
		if(parv[args] == NULL)
		{
		    break;
		}
		
		u = _TBL(user).get(parv[args]);
		g_return_if_fail(u != NULL);

		cm = find_chanmember(c, u);
		lp = find_userchannel(u, c);

		if(change == '+')
		{
		    cm->flags |= (*s == 'o' ? CHFL_CHANOP : CHFL_VOICE);
		    lp->flags |= (*s == 'o' ? MODE_CHANOP : MODE_VOICE);
		}
		else
		{
		    cm->flags &= ~((*s == 'o' ? CHFL_CHANOP : CHFL_VOICE));
		    lp->flags &= ~(*s == 'o' ? MODE_CHANOP : MODE_VOICE);
		}

		args++;

		break;

	    case 'b':
		/* ignore channel bans, for now
		 */
		break;

	    case 'l':
		if(change == '-')
		{
		    m->mode &= ~MODE_LIMIT;
		    m->limit = 0;

		    break;
		}
		else
		{
		    if(parv[args] == NULL)
		    {
			g_warning("Received a MODE %s %cl with no argument !", c->chname, change);
			break;
		    }

		    m->limit = atoi(parv[args]);
		    m->mode |= MODE_LIMIT;

		    args++;

		    break;
		}
		
	    case 'k':
		if(parv[args] == NULL)
		{
		    g_warning("Received a MODE %s %ck with no argument !", c->chname, change);
		    break;
		}

		if(change == '+')
		    g_strlcpy(m->key, parv[args], KEYLEN);
		else
		    *m->key = '\0';

		args++;

		break;

	    default:
		for(i = 1; flags[i] != 0; i += 2)
		{
		    if(*s == (gchar)flags[i])
		    {
			if(change == '+')
			    m->mode |= flags[i - 1];
			else
			    m->mode &= ~flags[i - 1];
			
			break;
		    }
		}

		break;
	}
	s++;
    }
}

/*
 * compile_umode()
 */
static G_INLINE_FUNC void compile_umode(User *u, gchar *s)
{
    gchar change = '+';
    gint i;
    static gint flags[] =
    {
	UMODE_o, 'o', UMODE_i, 'i', UMODE_r, 'r', UMODE_R, 'R',
	UMODE_y, 'y', UMODE_a, 'a', UMODE_A, 'A', UMODE_h, 'h',
	UMODE_x, 'x', UMODE_z, 'z', UMODE_S, 'S', 0, 0
    };

    g_return_if_fail(u != NULL);

    while(G_LIKELY(*s))
    {
	switch(*s)
	{
	    case '+':
	    case '-':
		change = *s;
		break;

	    default:
		for(i = 1; flags[i] != 0; i += 2)
		{
		    if(*s == (char)flags[i])
		    {
			if(change == '+')
			    u->mode |= flags[i - 1];
			else
			    u->mode &= ~flags[i - 1];

			break;
		    }
		}

		break;
	}
	s++;
    }
}


/* nick burst
 */
static void send_nick_burst(void)
{
    extern GSList *agents;
    GSList *johnny /* walker */;

    for(johnny = agents; johnny; johnny = g_slist_next(johnny))
    {
	agent_boot((Agent *)johnny->data);
    }
}

/* single entry point for PRIVMSG / NOTICE
 */
static G_INLINE_FUNC void m_message(User *u, gint parc, gchar **parv)
{
    gchar *srv_name;

    if(parc < 3)
	return;

    srv_name = strchr(parv[1], '@');
    if(srv_name)
    {
	*srv_name++ = '\0';
	if(mycmp(me.name, srv_name))
	{
	    g_warning("Fake direction, message for %s@%s arrived to me (%s)",
		    parv[1], srv_name, me.name);
	    return;
	}
    }

    my_strtoupper(parv[1]);
    agent_command(u, parv);
}

/* squit-related functions
 */
static gboolean add_squitted_server(GNode *n, User **sq_arr)
{
    gint i;

    for(i = 0; sq_arr[i] != NULL; i++);

    sq_arr[i] = (User *)n->data;

    return FALSE;
}

static gboolean delete_squitted_users(gchar *dummy, User *u, User **sq_arr)
{
    User *srv;
    gint i;

    if(u->flags & (FLAGS_SERVER|FLAGS_MYAGENT|FLAGS_MYCLIENT))
	return FALSE;

    for(i = 0; (srv = sq_arr[i]); i++)
    {
	if((u->server == srv))
	{
	    SLink *lp;

	    while(u->channels)
	    {
		lp = (SLink *)u->channels->data;
		g_return_val_if_fail(lp->v.c != NULL, FALSE);

		remove_user_from_channel(lp->v.c, u);
	    }

	    if(u->id_chans)
	    {
		g_slist_foreach(u->id_chans, (GFunc) free_char_slink, NULL);
		g_slist_free(u->id_chans);
	    }

	    if(u->id_nicks)
	    {
		g_slist_foreach(u->id_nicks, (GFunc) free_char_slink, NULL);
		g_slist_free(u->id_nicks);
	    }

	    return TRUE;
	}
    }

    return FALSE;
}

/* misc functions
 */
static void free_char_slink(gpointer data)
{
    g_free(((SLink *)data)->v.cp);
    g_mem_chunk_free(_MPL(links), data);
}

static gboolean dispatch_expire(void)
{
    extern GStaticMutex msg_mutex;

    g_static_mutex_lock(&msg_mutex);

    sql_lock_tables();
    message_dispatch(MESSAGE_EXPIRE, NULL, NULL);
    sql_unlock_tables();

    g_static_mutex_unlock(&msg_mutex);

    return TRUE;
}

/* EOF 
 */
