/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: modules.c,v 1.13 2003/03/27 01:30:37 vjt Exp $ */

#define MODULES
#include "modules.h"
#include "conf.h"
#include "main.h"
#include "log.h"
#include "match.h"

GSList *modules = NULL;
GStaticMutex msg_mutex = G_STATIC_MUTEX_INIT;

/* messages we recognize ..
 */
static GSList **messages;

static ModDesc *module_add_to_list(GModule *);
static gint module_find_callback(ModDesc *, gchar *);

static ModDesc *module_add_to_list(GModule *mod)
{
    gchar *name, *ver, *info;
    ModDesc *desc;

    g_return_val_if_fail(g_module_symbol(mod, "sux_module_name", (gpointer *) &name), NULL);
    g_return_val_if_fail(g_module_symbol(mod, "sux_module_version", (gpointer *) &ver), NULL);
    g_return_val_if_fail(g_module_symbol(mod, "sux_module_info", (gpointer *) &info), NULL);

    desc = g_new0(ModDesc, 1);
    desc->name = name;
    desc->version = ver;
    desc->info = info;
    desc->mod = mod;

    modules = g_slist_prepend(modules, (gpointer) desc);

    return desc;
}

void load_modules(void)
{
    GNode *modules_section;
    guint loaded_modules = 0;
    gchar *modstring;

    messages = g_new0(GSList *, MESSAGE_LAST);
    modules_section = find_section_id(SYMBOL_SECTION_MODULES);
    if(!modules_section)
    {
	g_warning("Cannot find modules section !");
	return;
    }

    if((modstring = find_conf_id(modules_section, SYMBOL_MODULE_LOAD)))
    {
	gchar **modules, *module;
	gint i;
	
	modules = g_strsplit(modstring, ";", -1);
	for(i = 0; modules[i]; i++)
	{
	    module = g_strstrip(modules[i]);
	    if(module_load(module))
	    {
		loaded_modules++;
	    }
	}

	g_strfreev(modules);
    }

    g_message_syslog("MDL: Loaded %d module%s, firing up !",
	    loaded_modules, loaded_modules == 1 ? "" : "s");

    return;
}

ModDesc *module_load(gchar *modname)
{
    extern options_t options;
    gchar *modfile, *modpath;
    GModule *mod;
    ModDesc *ret;

    modfile = g_strrstr(modname, ".so") ?
	g_strdup(modname) :
	g_strdup_printf("%s.so", modname);

    modpath = g_build_filename(options.moddir, modfile, NULL);
    mod = g_module_open(modpath, 0);

    if(!mod)
    {
	g_warning("Cannot load `%s' module: %s",
		modname, g_module_error());

	ret = NULL;
    }
    else
    {
	ret = module_add_to_list(mod);
    }

    g_free(modpath);
    g_free(modfile);

    return ret;
}

gboolean module_unload(gchar *modname)
{
    GSList *loaded_mod;
    ModDesc *mod;
    
    loaded_mod = g_slist_find_custom(modules, modname, (GCompareFunc) module_find_callback);
    if(!loaded_mod)
    {
	return FALSE;
    }

    mod = (ModDesc *)loaded_mod->data;
    if(!g_module_close(mod->mod))
    {
	g_warning("Cannot unload `%s' module: %s",
		mod->name, g_module_error());
	return FALSE;
    }

    modules = g_slist_remove_link(modules, loaded_mod);
    g_free(loaded_mod->data);
    g_slist_free_1(loaded_mod);

    return TRUE;
}

ModDesc *module_find(gchar *modname)
{
    GSList *loaded_mod = g_slist_find_custom(modules, modname, (GCompareFunc) module_find_callback);

    return (loaded_mod) ? ((ModDesc *) loaded_mod->data) : NULL;
}

static gint module_find_callback(ModDesc *haystack, gchar *needle)
{
    return (mycmp(haystack->name, needle));
}

static gint sux_message_priority_compare_callback(gconstpointer a, gconstpointer b)
{
    gint new_msg_pri = ((SuxMessage *)a)->priority, old_msg_pri = ((SuxMessage *)b)->priority;
    
    if(new_msg_pri > old_msg_pri)
    {
	return -1;
    }
    else if(old_msg_pri > new_msg_pri)
    {
	return 1;
    }
    else
    {
	return 0;
    }
}

static gint sux_message_find(gconstpointer a, gconstpointer b)
{
    return((((SuxMessage *)a)->callback == b) ? 0 : -1);
}

/* message functions
 */
void message_add_callback(gint id, SuxMsgPriority priority, message_func callback)
{
    SuxMessage *msg = g_new0(SuxMessage, 1);

    msg->priority = priority;
    msg->callback = callback;

    g_static_mutex_lock(&msg_mutex);
    messages[id] = g_slist_insert_sorted(messages[id], msg, sux_message_priority_compare_callback);
    g_static_mutex_unlock(&msg_mutex);
}

void message_del_callback(gint id, message_func callback)
{
    GSList *message = g_slist_find_custom(messages[id], callback, sux_message_find);

    if(message == NULL)
    {
	g_warning("callback %p not found in message id %d !",
		callback, id);
	return;
    }

    g_static_mutex_lock(&msg_mutex);
    messages[id] = g_slist_remove_link(messages[id], message);
    g_static_mutex_unlock(&msg_mutex);

    g_free(message->data);
    g_slist_free_1(message);
}

G_INLINE_FUNC void message_dispatch(gint id, gpointer data1, gpointer data2)
{
    register GSList *johnny = messages[id];

    while(johnny && (*((SuxMessage *)johnny->data)->callback)(data1, data2))
    {
	johnny = g_slist_next(johnny);
    }
}
