/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: nickserv.c,v 1.49 2003/03/29 01:58:55 vjt Exp $ */

#include "sux.h"
#include "agents.h"
#include "nickserv.h"
#include "sql.h"
#include "main.h"
#include "network.h"
#include "match.h"
#include "help.h"
#include "modules.h"
#include "threads.h"
#include "sux-printf.h"
#include "conf.h"
#include "ns-val-email.h"

User *ns;
static MYSQL_RES *ns_res;
static MYSQL_ROW ns_row;

REMOTE_TABLE_INSTANCE(user);
REMOTE_MEMPOOL_INSTANCE(links);

const gchar sux_module_name[] = "NickServ";
const gchar sux_module_version[] = "0.1.3";
const gchar sux_module_info[] = "Nickname Management Services";

/* synch
 */
static void ns_initial_synch(void);
static void synch_user(gchar *, User *);

/* id-nick related
 */
static G_INLINE_FUNC SLink *ns_find_id_nick(User *, gchar *);
static G_INLINE_FUNC SLink *ns_create_id_nick(gchar *);
static G_INLINE_FUNC void ns_add_id_nick(User *, gchar *);
static void ns_check_reg_nick(User *);

/* enforcers
 */
static void ns_create_enforcer(gchar *);
static gboolean ns_enforce(User *);
static gboolean ns_enforce_15sec(User *);
static gboolean ns_enforce_30sec(User *);
static gboolean ns_enforce_45sec(User *);
static gboolean ns_quit_enforcer(User *);
static void ns_delete_enforcers(void);
static gboolean ns_delete_enforcers_callback(gchar *, User *);

/* messages
 */
static gboolean ns_message_newnick_callback(User *);
static gboolean ns_message_nickchange_callback(User *, gchar *);

/* nick expire
 */
static gboolean ns_expire_nicknames(void);

/* module init
 */
gchar *g_module_check_init(GModule *mod)
{
    extern GSList *agents;
    Agent *agent;

    agent = g_new0(Agent, 1);
    agent->name = "NickServ";
    agent->id = SYMBOL_SECTION_NICKSERV;
    agent->u = &ns;

    agents = g_slist_prepend(agents, (gpointer) agent);

    agent_setcallback("NICKSERV", nickserv);

    ns_init_validate_email();

    message_add_callback(MESSAGE_NEWNICK, SUX_MSG_PRIORITY_LOW, (message_func) ns_message_newnick_callback);
    message_add_callback(MESSAGE_NICKCHANGE, SUX_MSG_PRIORITY_NORMAL, (message_func) ns_message_nickchange_callback);
    message_add_callback(MESSAGE_EXPIRE, SUX_MSG_PRIORITY_NORMAL, (message_func) ns_expire_nicknames);

    message_add_callback(MESSAGE_NSREQ, SUX_MSG_PRIORITY_NORMAL, (message_func) nickserv);

    if(me.boot)
    {
	agent_boot(agent);
	ns_initial_synch();
    }

    g_message_syslog("MDL: Loaded NickServ module");

    return NULL;
}

/* module cleanup
 */
void g_module_unload(GModule *mod)
{
    extern GSList *agents;
    GSList *agent;

    agent = g_slist_find_custom(agents, GINT_TO_POINTER(SYMBOL_SECTION_NICKSERV),
	    (GCompareFunc) agent_find);
    agents = g_slist_remove_link(agents, agent);
    agent_shutdown((Agent *)agent->data);

    ns_cleanup_validate_email();
    ns_delete_enforcers();

    g_free(agent->data);
    g_slist_free_1(agent);

    message_del_callback(MESSAGE_NEWNICK, (message_func) ns_message_newnick_callback);
    message_del_callback(MESSAGE_NICKCHANGE, (message_func) ns_message_nickchange_callback);
    message_del_callback(MESSAGE_EXPIRE, (message_func) ns_expire_nicknames);

    message_del_callback(MESSAGE_NSREQ, (message_func) nickserv);

    g_message_syslog("MDL: Unloaded NickServ module");
}

void ns_help(User *caller, gint parc, gchar **parv)
{
    sux_help(caller, ns, NICKSERV_HELP_ID, parv ? parv[0] : NULL);
}

/*
 * parv[0] -> password
 * parv[1] -> email
 */
void ns_register(User *caller, gint parc, gchar **parv)
{
    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		ns->name, caller->name);
	return;
    }

    if(myncmp(caller->name, "Guest", 5) == 0)
    {
	send_out(":%s NOTICE %s :Guest nicks cannot be registered or used",
		ns->name, caller->name);
	return;
    }

    if(strlen(parv[1]) >= EMAILLEN)
    {
	send_out(":%s NOTICE %s :Too long email address",
		ns->name, caller->name);
	return;
    }
	
    if(!ns_validate_email(parv[1]))
    {
	send_out(":%s NOTICE %s :'%s': invalid email address",
		ns->name, caller->name, parv[1]);
	return;
    }

    sql_query_assert(&ns_res,
	    "SELECT `name` FROM `nicks` WHERE `name` = '%s'",
	    caller->name);

    if(mysql_num_rows(ns_res))
    {
	send_out(":%s NOTICE %s :Nick %s is already registered",
		ns->name, caller->name, caller->name);
	
	mysql_free_result(ns_res);
	return;
    }

    mysql_free_result(ns_res);

    /* ok, proceed to registration
     */

    sql_query_assert(NULL,
	    "INSERT INTO `nicks` VALUES(0, '%s', '%s', NOW(), NOW(), NULL, NULL, '%s', NULL, NULL, NULL)",
	    caller->name, parv[0], parv[1]);

    send_out(":%s NOTICE %s :Nick %s is now registered to you. Your password is %s",
	    ns->name, caller->name, caller->name, parv[0]);

    send_out("SVSMODE %s %ld :+r", caller->name, caller->ts);
    caller->mode |= UMODE_r;

    caller->id_nicks = g_slist_prepend(caller->id_nicks, ns_create_id_nick(caller->name));

    return;
}

/*
 * parv[0] -> password
 */
void ns_identify(User *caller, gint parc, gchar **parv)
{
    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		ns->name, caller->name);
	return;
    }

    if(myncmp(caller->name, "Guest", 5) == 0)
    {
	send_out(":%s NOTICE %s :Guest nicks cannot be registered or used",
		ns->name, caller->name);
	return;
    }

    if(ns_find_id_nick(caller, caller->name))
    {
	send_out(":%s NOTICE %s :You are already identified to nick %s",
		ns->name, caller->name, caller->name);
	return;
    }

    sql_query_assert(&ns_res,
	    "SELECT `pass` FROM `nicks` WHERE `name` = '%s'",
	    caller->name);

    if(!mysql_num_rows(ns_res))
    {
	send_out(":%s NOTICE %s :Nick %s is not registered",
		ns->name, caller->name, caller->name);
	mysql_free_result(ns_res);

	return;
    }

    ns_row = mysql_fetch_row(ns_res);
    if(strcmp(*ns_row, parv[0]))
    {
	send_out(":%s NOTICE %s :Password supplied for %s is incorrect",
		ns->name, caller->name, caller->name);
	mysql_free_result(ns_res);
	return;
    }

    if(caller->id_timeout)
    {
	g_source_del(caller->id_timeout);
	caller->id_timeout = NULL;
    }

    mysql_free_result(ns_res);

    sql_query_assert(NULL,
	    "UPDATE `nicks` SET `reg_time` = `reg_time`, `last_seen` = NOW() "
	    "WHERE `name` = '%s'",
	    caller->name);

    send_out(":%s NOTICE %s :Password accepted for %s, you are now identified",
	    ns->name, caller->name, caller->name);
    send_out("SVSMODE %s %ld :+r", caller->name, caller->ts);
    caller->mode |= UMODE_r;

    caller->id_nicks = g_slist_prepend(caller->id_nicks, ns_create_id_nick(caller->name));

    message_dispatch(MESSAGE_IDENTIFY, caller, NULL);
    
    return;
}

/*
 * parv[0] -> nick to query
 */
void ns_info(User *caller, gint parc, gchar **parv)
{
    User *u;

    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		ns->name, caller->name);

	return;
    }

    /* ctime format: "Wed Jun 30 21:49:08 1993\n"
     */
    sql_query_assert(&ns_res,
	    "SELECT `name`, `pass`, DATE_FORMAT(`reg_time`, '%%b %%e %%Y - %%T'), "
	    "DATE_FORMAT(`last_seen`, '%%b %%e %%Y - %%T'), "
	    "DATE_FORMAT(NOW(), '%%b %%e %%Y - %%T'), "
	    "`flags`, `url`, `email`, `hold`, "
	    "`mark`, `forbid` FROM `nicks` WHERE `name` = '%s'",
	    parv[0]);

    if(!mysql_num_rows(ns_res))
    {
	send_out(":%s NOTICE %s :Nick %s is not registered",
		ns->name, caller->name, parv[0]);
	mysql_free_result(ns_res);

	return;
    }

    ns_row = mysql_fetch_row(ns_res);
    u = _TBL(user).get(parv[0]);
    
    send_out(":%s NOTICE %s :*** Info about %s", ns->name, caller->name, parv[0]);
    if(u)
    {
	if(u->mode & UMODE_r)
	{
	    send_out(":%s NOTICE %s :Real name: %s", ns->name, caller->name, u->info);
	    send_out(":%s NOTICE %s :Address: %s", ns->name, caller->name, u->host);
	    send_out(":%s NOTICE %s :Server: %s", ns->name, caller->name, caller->server->name);
	    send_out(":%s NOTICE %s :[Currently on IRC] [Identified]", ns->name, caller->name);
	}
	else
	{
	    send_out(":%s NOTICE %s :[Currently on IRC] [Not Identified]", ns->name, caller->name);
	}
    }
    
    send_out(":%s NOTICE %s :Registration time: %s", ns->name, caller->name, ns_row[2]);
    send_out(":%s NOTICE %s :Last seen: %s", ns->name, caller->name, ns_row[3]);
    send_out(":%s NOTICE %s :Time now: %s", ns->name, caller->name, ns_row[4]);
    send_out(":%s NOTICE %s :E-Mail: %s", ns->name, caller->name, ns_row[7]);

    /* XXX: hold, forbid, etc
     */

    send_out(":%s NOTICE %s :*** End of Info ***", ns->name, caller->name);

    mysql_free_result(ns_res);
    return;
}

void ns_ghost(User *caller, gint parc, gchar **parv)
{
    gchar *nick, *pass;
    User *ghosted;

    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		ns->name, caller->name);
	return;
    }

    nick = parv[0];
    pass = parv[1];
    ghosted = _TBL(user).get(nick);

    if(!ghosted)
    {
	send_out(":%s NOTICE %s :Nick %s is not online",
		ns->name, caller->name, nick);
	return;
    }

    if(ghosted->flags & FLAGS_MYCLIENT)
    {
	send_out(":%s NOTICE %s :%s is a services enforcer, use /ns RELEASE",
		ns->name, caller->name, nick);
	return;
    }

    if(ghosted->flags & (FLAGS_MYAGENT|FLAGS_SERVER))
    {
	return;
    }

    if(!ns_find_id_nick(caller, nick))
    {
	sql_query_assert(&ns_res,
		"SELECT `pass` FROM `nicks` WHERE `name` = '%s'",
		nick);

	if(!mysql_num_rows(ns_res))
	{
	    send_out(":%s NOTICE %s :Nick %s is not registered",
		    ns->name, caller->name, nick);
	    mysql_free_result(ns_res);

	    return;
	}

	ns_row = mysql_fetch_row(ns_res);
	if(strcmp(*ns_row, pass))
	{
	    send_out(":%s NOTICE %s :Password supplied for %s is incorrect",
		    ns->name, caller->name, nick);
	    mysql_free_result(ns_res);
	    return;
	}
	mysql_free_result(ns_res);
    }

    /* ok, ghost the user
     */
    send_out(":%s KILL %s :%s (GHOST Command used by %s)",
	    me.name, ghosted->name, me.name, caller->name);
    exit_client(ghosted);

    send_out(":%s NOTICE %s :Your ghost has been disconnected",
	    ns->name, caller->name);
}

void ns_release(User *caller, gint parc, gchar **parv)
{
    gchar *nick, *pass;
    User *ghosted;

    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		ns->name, caller->name);
	return;
    }

    nick = parv[0];
    pass = parv[1];
    ghosted = _TBL(user).get(nick);

    if(!ghosted)
    {
	send_out(":%s NOTICE %s :Nick %s is not online",
		ns->name, caller->name, nick);
	return;
    }

    if(!(ghosted->flags & FLAGS_MYCLIENT))
    {
	send_out(":%s NOTICE %s :%s is not services enforcer, use /ns GHOST",
		ns->name, caller->name, nick);
	return;
    }

    if(ghosted->flags & (FLAGS_MYAGENT|FLAGS_SERVER))
    {
	return;
    }

    if(!ns_find_id_nick(caller, nick))
    {
	sql_query_assert(&ns_res,
		"SELECT `pass` FROM `nicks` WHERE `name` = '%s'",
		nick);

	if(!mysql_num_rows(ns_res))
	{
	    send_out(":%s NOTICE %s :Nick %s is not registered",
		    ns->name, caller->name, nick);
	    mysql_free_result(ns_res);

	    return;
	}

	ns_row = mysql_fetch_row(ns_res);
	if(strcmp(*ns_row, pass))
	{
	    send_out(":%s NOTICE %s :Password supplied for %s is incorrect",
		    ns->name, caller->name, nick);
	    mysql_free_result(ns_res);
	    return;
	}
	mysql_free_result(ns_res);
    }

    /* ok, quit the enforcer
     */
    send_out(":%s QUIT :RELEASE Command used by %s",
	    ghosted->name, caller->name);
    exit_client(ghosted);

    send_out(":%s NOTICE %s :The enforcer has been disconnected",
	    ns->name, caller->name);
}

void ns_recover(User *caller, gint parc, gchar **parv)
{
    gchar *nick, *pass;
    User *ghosted;
    guint32 rnd;
    gchar newnick[NICKLEN + 1];

    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		ns->name, caller->name);
	return;
    }

    nick = parv[0];
    pass = parv[1];
    ghosted = _TBL(user).get(nick);

    if(!ghosted)
    {
	send_out(":%s NOTICE %s :Nick %s is not online",
		ns->name, caller->name, nick);
	return;
    }

    if(caller == ghosted)
    {
	send_out(":%s NOTICE %s :You cannot use RECOVER on yourself",
		ns->name, caller->name);
	return;
    }

    if(ghosted->flags & FLAGS_MYCLIENT)
    {
	send_out(":%s NOTICE %s :%s is a services enforcer, use /ns RELEASE",
		ns->name, caller->name, nick);
	return;
    }

    if(ghosted->flags & (FLAGS_MYAGENT|FLAGS_SERVER))
    {
	return;
    }

    if(!ns_find_id_nick(caller, nick))
    {
	sql_query_assert(&ns_res,
		"SELECT `pass` FROM `nicks` WHERE `name` = '%s'",
		nick);

	if(!mysql_num_rows(ns_res))
	{
	    send_out(":%s NOTICE %s :Nick %s is not registered",
		    ns->name, caller->name, nick);
	    mysql_free_result(ns_res);

	    return;
	}

	ns_row = mysql_fetch_row(ns_res);
	if(strcmp(*ns_row, pass))
	{
	    send_out(":%s NOTICE %s :Password supplied for %s is incorrect",
		    ns->name, caller->name, nick);
	    mysql_free_result(ns_res);
	    return;
	}
	mysql_free_result(ns_res);

	ns_add_id_nick(caller, nick);
    }

    /* ok, recover the user
     */
    do
    {
	rnd = g_random_int_range(0x0, 0xffff);
	sux_snprintf(newnick, sizeof(newnick), "Guest%d", rnd);
    } while(_TBL(user).get(newnick));

    send_out("SVSNICK %s %s %ld", ghosted->name, newnick, NOW);

    send_out(":%s NOTICE %s :Your nick has been recovered",
	    ns->name, caller->name);
}

static G_INLINE_FUNC SLink *ns_find_id_nick(User *u, gchar *arnold)
{
    register GSList *johnny = u->id_nicks; /* walker */

    while(johnny)
    {
	if(!mycmp(((SLink *)johnny->data)->v.cp, arnold)) /* search terminated */
	{
	    return ((SLink *) johnny->data);
	}

	johnny = g_slist_next(johnny);
    }

    return NULL;
}

static G_INLINE_FUNC SLink *ns_create_id_nick(gchar *nick)
{
    SLink *ret;

    ret = g_mem_chunk_alloc0(_MPL(links));
    ret->v.cp = g_strdup(nick);

    return ret;
}

static G_INLINE_FUNC void ns_add_id_nick(User *u, gchar *nick)
{
    if(!ns_find_id_nick(u, nick))
    {
	u->id_nicks = g_slist_prepend(u->id_nicks,
		ns_create_id_nick(nick));
    }
}

static void ns_check_reg_nick(User *u)
{
    if(u->mode & UMODE_r || ns_find_id_nick(u, u->name))
    {
	if(!(u->mode & UMODE_r))
	{
	    send_out("SVSMODE %s %ld :+r", u->name, u->ts);
	    u->mode |= UMODE_r;
	}

	return;
    }

    sql_query_assert(&ns_res,
	    "SELECT `id` FROM `nicks` WHERE `name` = '%s'",
	    u->name);

    if(mysql_num_rows(ns_res))
    {	
	send_out(":%s NOTICE %s :This nickname is owned by someone else",
		ns->name, u->name);
	send_out(":%s NOTICE %s :If this is your nickname, type /ns IDENTIFY <password>",
		ns->name, u->name);
	send_out(":%s NOTICE %s :Your nick will be changed in 60 seconds if you do not comply",
		ns->name, u->name);
	u->id_timeout = g_timeout_source_add(15 gs_seconds, (GSourceFunc) ns_enforce_45sec, u);
    }

    mysql_free_result(ns_res);
}

static void ns_create_enforcer(gchar *nick)
{
    User *u = _TBL(user).alloc(nick);

    strcpy(u->username, "enforcer");
    strcpy(u->host, "enforcer.service");
    strcpy(u->info, "Nick Protection Enforcement");
    u->server = me.table_ptr;
    u->mode = UMODE_i;
    u->ts = NOW;
    u->flags |= FLAGS_MYCLIENT;
    u->id_timeout = g_timeout_source_add(1 gs_minute, (GSourceFunc) ns_quit_enforcer, u);

    client_introduce(u);
}

static gboolean ns_quit_enforcer(User *u)
{
    send_out(":%s QUIT :Enforcer quitting", u->name);

    if(u->id_timeout)
    {
	g_source_del(u->id_timeout);
    }

    _TBL(user).del(u);

    return FALSE;
} 

static gboolean ns_delete_enforcers_callback(gchar *dummy, User *u)
{
    if(u->id_timeout)
    {
	g_source_del(u->id_timeout);
	u->id_timeout = NULL;
    }

    if(u->flags & FLAGS_MYCLIENT)
    {
	send_out(":%s QUIT :Enforcer quitting due to NickServ unload",
		u->name);
	return TRUE;
    }

    return FALSE;
}

static void ns_delete_enforcers(void)
{
    g_hash_table_foreach_remove(_TBL(user).__hash_tbl,
	    (GHRFunc) ns_delete_enforcers_callback, NULL);
}

static gboolean ns_enforce_45sec(User *u)
{
    send_out(":%s NOTICE %s :Your nick will be changed in 45 seconds if you do not comply",
	    ns->name, u->name);

    g_source_del(u->id_timeout);

    u->id_timeout = g_timeout_source_add(15 gs_seconds, (GSourceFunc) ns_enforce_30sec, u);

    return FALSE;
}

static gboolean ns_enforce_30sec(User *u)
{
    send_out(":%s NOTICE %s :Your nick will be changed in 30 seconds if you do not comply",
	    ns->name, u->name);

    g_source_del(u->id_timeout);

    u->id_timeout = g_timeout_source_add(15 gs_seconds, (GSourceFunc) ns_enforce_15sec, u);

    return FALSE;
}

static gboolean ns_enforce_15sec(User *u)
{
    send_out(":%s NOTICE %s :Your nick will be changed in 15 seconds if you do not comply",
	    ns->name, u->name);

    g_source_del(u->id_timeout);

    u->id_timeout = g_timeout_source_add(15 gs_seconds, (GSourceFunc) ns_enforce, u);

    return FALSE;
}

static gboolean ns_enforce(User *u)
{
    guint32 rnd;
    gchar newnick[NICKLEN + 1];

    send_out(":%s NOTICE %s :Your nick will now be changed",
	    ns->name, u->name);

    do
    {
	rnd = g_random_int_range(0x0, 0xffff);
	sux_snprintf(newnick, sizeof(newnick), "Guest%d", rnd);
    } while(_TBL(user).get(newnick));
	
    send_out("SVSNICK %s %s %ld", u->name, newnick, NOW);
    u->flags |= FLAGS_ENFORCE;

    g_source_del(u->id_timeout);
    u->id_timeout = NULL;

    return FALSE;
}

/*
 * synch
 */
static void ns_initial_synch(void)
{
    time_t synch_time;

    sql_lock_tables();

    synch_time = time(NULL);
    send_out(":%s GLOBOPS :Starting user synch with database", me.name);
    g_hash_table_foreach(_TBL(user).__hash_tbl, (GHFunc) synch_user, NULL);
    send_out(":%s GLOBOPS :NickServ synch completed in %ld second(s)", me.name, time(NULL) - synch_time);

    sql_unlock_tables();
}

static void synch_user(gchar *dummy, User *u)
{
    if(u->flags & (FLAGS_MYAGENT|FLAGS_MYCLIENT|FLAGS_SERVER))
    {
	return;
    }

    if(!(u->mode & UMODE_r))
    {
	ns_check_reg_nick(u);
    }
}

/* 
 * messages callbacks
 */
static gboolean ns_message_newnick_callback(User *u)
{
    if(u->mode & UMODE_r)
    {
	/* user was previously identified to this nick,
	 * trust it and add the nick to the list of the
	 * id nicks
	 */
	ns_add_id_nick(u, u->name);
    }
    else
    {
	ns_check_reg_nick(u);
    }

    return TRUE;
}

static gboolean ns_message_nickchange_callback(User *u, gchar *oldnick)
{
    if(u->flags & FLAGS_ENFORCE)
    {
	ns_create_enforcer(oldnick);
	u->flags &= ~FLAGS_ENFORCE;
	u->mode &= ~UMODE_r;
    }
    else
    {
	if(ns_find_id_nick(u, u->name))
	{
	    send_out("SVSMODE %s %ld :+r", u->name, u->ts);
	    u->mode |= UMODE_r;
	}
	else if(u->mode & UMODE_r)
	{
	    u->mode &= ~UMODE_r;
	}

	if(!(u->mode & UMODE_r))
	{
	    ns_check_reg_nick(u);
	}
    }

    return TRUE;
}

/* nickname expiration
 */
static gboolean ns_expire_nicknames(void)
{
    static guint expire_no = 0;
    MYSQL_RES *expired_channels_res;
    MYSQL_ROW expired_channels_row;
    time_t expire_started = time(NULL);
    guint deleted_nicks = 0, deleted_chans = 0, deleted_memos = 0;

    send_out(":%s GLOBOPS :Starting nickname expire #%d",
	    ns->name, ++expire_no);
    
    sql_query_assert_ret(FALSE, &ns_res, "SELECT `id`, `name` FROM `nicks` "
	    "WHERE TO_DAYS(NOW()) - TO_DAYS(`last_seen`) > 40");

    while((ns_row = mysql_fetch_row(ns_res)))
    {
	gchar *nick_id = ns_row[0];

	send_out(":%s PRIVMSG #sux :Expiring nickname %s",
		ns->name, ns_row[1]);

	sql_query_assert_ret(FALSE, &expired_channels_res, "SELECT `id`, `name` FROM `chans` "
		    "WHERE `founder_id` = %s",
		    nick_id);
	
	while((expired_channels_row = mysql_fetch_row(expired_channels_res)))
	{
	    gchar *chan_id = expired_channels_row[0];
	    const gchar *channel_tables[] = {
		"akick",
		"aop",
		"cfounder",
		"sop",
		"vop",
		NULL
	    };
	    gint i;

	    send_out(":%s PRIVMSG #sux :Expiring channel %s ( cascade for previous nick expiration )",
		    ns->name, expired_channels_row[1]);

	    for(i = 0; channel_tables[i]; i++)
	    {
		sql_query_assert_ret(FALSE, NULL, "DELETE FROM `%s` WHERE `chan_id` = %s OR `nick_id` = %s",
		    channel_tables[i], chan_id, nick_id);
	    }
	}

	mysql_free_result(expired_channels_res);
	sql_query_assert_ret(FALSE, NULL, "DELETE FROM `chans` WHERE `founder_id` = %s",
		nick_id);
	deleted_chans += sql_affected_rows();

	sql_query_assert_ret(FALSE, NULL, "DELETE FROM `memo` WHERE `rcpt_id` = %s",
		nick_id);
	deleted_memos += sql_affected_rows();
    }

    mysql_free_result(ns_res);
    sql_query_assert_ret(FALSE, NULL, "DELETE FROM `nicks` WHERE TO_DAYS(NOW()) - TO_DAYS(`last_seen`) > 40");
    deleted_nicks = sql_affected_rows();

    send_out(":%s GLOBOPS :Nickname expire completed in %ld secs. "
	    "Deleted %d nicknames, %d channels ( cascade ), %d memos ( cascade )",
	    ns->name, time(NULL) - expire_started, deleted_nicks, deleted_chans, deleted_memos);

    return TRUE;
}
