/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: operserv-akill.c,v 1.12 2003/03/27 02:10:29 vjt Exp $ */

#include "sux.h"
#include "network.h"
#include "parse.h"
#include "sql.h"
#include "main.h"
#include "match.h"
#include "threads.h"
#include "sux-printf.h"
#include "operserv.h"

/* variables
 */
extern User *os;
extern GMutex *os_akill_mutex;
extern GSList *active_akills;

static MYSQL_RES *os_akill_res;
static MYSQL_ROW os_akill_row;


/* private functions
 */
static G_INLINE_FUNC void os_akill_add_internal(User *, gchar **, gint);
static gint os_akill_parse_time(gchar *);
static gboolean os_expire_akill(AKill *);
static gint akill_find(AKill *, gint);
static gchar *pretty_expiration_time(time_t, gchar *, gsize);

void os_akill_add(User *caller, gint parc, gchar **parv)
{
    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    /* fixed 3-hours expiry
     */
    os_akill_add_internal(caller, parv, 3600 * 3);
}

void os_akill_del(User *caller, gint parc, gchar **parv)
{
    gchar sqlmask[SQLMASKLEN + 1];
    gchar *id, *user, *host;
    GSList *l;

    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    if(strchr(parv[0], '@'))
    {
	sql_query_assert(&os_akill_res, "SELECT `id`, `display_mask` FROM `akill` WHERE `mask` = '%s'",
		match_to_sqlmask(parv[0], sqlmask, SQLMASKLEN));
    }
    else
    {
	sql_query_assert(&os_akill_res, "SELECT `id`, `display_mask` FROM `akill` WHERE `id` = '%s'",
		parv[0]);
    }

    if(!mysql_num_rows(os_akill_res))
    {
	send_out(":%s NOTICE %s :%s was not found on the akill list",
		os->name, caller->name, parv[0]);

	mysql_free_result(os_akill_res);

	return;
    }

    os_akill_row = mysql_fetch_row(os_akill_res);
    id = os_akill_row[0];
    user = os_akill_row[1];
    host = strchr(os_akill_row[1], '@');

    if(!host)
    {
	g_critical("Badly formatted akill in DB: %s (ID: %s)",
		os_akill_row[1], os_akill_row[0]);
	return;
    }

    *host++ = '\0';

    sql_query_assert(NULL, "DELETE FROM `akill` WHERE `id` = '%s'", id);

    g_mutex_lock(os_akill_mutex);
    l = g_slist_find_custom(active_akills, GINT_TO_POINTER(atoi(id)), (GCompareFunc) akill_find);
    if(l)
    {
	AKill *a = l->data;
	g_source_del(a->timeout);
	active_akills = g_slist_remove(active_akills, a);

	g_free(a->user);
	g_free(a->host);

	g_free(a);
    }
    g_mutex_unlock(os_akill_mutex);
    
    send_out("RAKILL %s %s", host, user);
    send_out(":%s NOTICE %s :AKILL for %s@%s removed",
	    os->name, caller->name, user, host);
    send_out(":%s GLOBOPS :%s (%s@%s) removed an akill for %s@%s (ID: %s)",
	    os->name, caller->name, caller->username, caller->host, user, host, id);

    mysql_free_result(os_akill_res);

    return;
}

void os_akill_list(User *caller, gint parc, gchar **parv)
{
    gint i = 0;
    GSList *johnny /* walker */;

    if(parc)
    {
	gchar sqlmask[SQLMASKLEN + 1];
	send_out(":%s NOTICE %s :*** Permanent AutoKill list (matching %s) ***",
		os->name, caller->name, parv[0]);
    
	match_to_sqlmask(parv[0], sqlmask, SQLMASKLEN);
	sql_query_assert(&os_akill_res, "SELECT `id`, `display_mask`, `reason`, `set_by`, "
		"DATE_FORMAT(`time_set`, '%%b %%e %%Y - %%T') "
		"FROM `akill` "
		"WHERE `expire` = 0 AND `display_mask` LIKE '%s' "
		"ORDER BY `id`", sqlmask);
    }
    else
    {
	send_out(":%s NOTICE %s :*** Permanent AutoKill list ***",
		os->name, caller->name);
    
	sql_query_assert(&os_akill_res, "SELECT `id`, `display_mask`, `reason`, `set_by`, "
		"DATE_FORMAT(`time_set`, '%%b %%e %%Y - %%T') "
		"FROM `akill` "
		"WHERE `expire` = 0 "
		"ORDER BY `id`");
    }

    while((os_akill_row = mysql_fetch_row(os_akill_res)))
    {
	/* +----+--------------+--------+--------+----------+--------+
	 * | id | display_mask | reason | set_by | time_set | expire |
	 * +----+--------------+--------+--------+----------+--------+
	 */
	send_out(":%s NOTICE %s :%s) %s - Set by %s on %s. Reason: %s",
		os->name,        /**/
		caller->name,    /*  */
		os_akill_row[0], /* ID */
		os_akill_row[1], /* mask */
		os_akill_row[3], /* set by */
		os_akill_row[4], /* time set */
		os_akill_row[2]) /* ban reason */;
	++i;
    }

    send_out(":%s NOTICE %s :*** End of list. Entries: %d ***",
	    os->name, caller->name, i);

    g_mutex_lock(os_akill_mutex);
    johnny = active_akills;
    if(johnny)
    {
	i = 0;
	send_out(":%s NOTICE %s :*** Temporary AutoKill List ***",
		os->name, caller->name);
	while(johnny)
	{
	    AKill *a = johnny->data;
	    gchar expiration_time[128];

	    pretty_expiration_time(a->expire_at - NOW, expiration_time, sizeof(expiration_time));
	    send_out(":%s NOTICE %s :%d) %s@%s, expires in %s",
		    os->name, caller->name, a->id, a->user, a->host, expiration_time);

	    johnny = g_slist_next(johnny);
	    ++i;
	}
	send_out(":%s NOTICE %s :*** End of list. Entries: %d ***",
		os->name, caller->name, i);
    }
    g_mutex_unlock(os_akill_mutex);
}

void os_akill_perm(User *caller, gint parc, gchar **parv)
{
    if(parc < 2)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    os_akill_add_internal(caller, parv, 0);
}

void os_akill_time(User *caller, gint parc, gchar **parv)
{
    gint expiration_time;
    
    if(parc < 3)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    expiration_time = os_akill_parse_time(parv[0]);

    if(expiration_time < 0)
    {
	send_out(":%s NOTICE %s :Invalid time format (%s)",
		os->name, caller->name, parv[0]);

	return;
    }

    os_akill_add_internal(caller, &parv[1], expiration_time);
}

void os_akill_id(User *caller, gint parc, gchar **parv)
{
    if(parc < 1)
    {
	send_out(":%s NOTICE %s :Not enough parameters",
		os->name, caller->name);
	return;
    }

    sql_query_assert(&os_akill_res,
	    "SELECT `id`, `display_mask`, `reason`, `set_by`, UNIX_TIMESTAMP(`time_set`), UNIX_TIMESTAMP(`expire`) "
	    "FROM `akill` "
	    "WHERE `id` = %s",
	    parv[0]);

    if(!mysql_num_rows(os_akill_res))
    {
	mysql_free_result(os_akill_res);
	send_out(":%s NOTICE %s :Non-existant AKILL ID %s",
		os->name, caller->name, parv[0]);
	return;
    }

    os_akill_row = mysql_fetch_row(os_akill_res);
    
    send_out(":%s NOTICE %s :Showing AKILL ID %s",
	    os->name, caller->name, parv[0]);
    send_out(":%s NOTICE %s :1) %s - Set by: %s on: %s; "
	    "expires: %s; reason: '%s' (ID: %s)",
	    os->name, caller->name, os_akill_row[1], os_akill_row[3],
	    os_akill_row[4], os_akill_row[5], os_akill_row[2], os_akill_row[0]);

    return;
}

void os_akill_db_synch(void)
{
    /* Delete old akills
     */
    sql_query_assert(NULL,
	    "DELETE FROM `akill` "
	    "WHERE `expire` != 0 AND `expire` <= NOW()");

    /* Load temporary akills
     */
    sql_query_assert(&os_akill_res,
	    "SELECT `id`, `display_mask`, UNIX_TIMESTAMP(`expire`) "
	    "FROM `akill` "
	    "WHERE `expire` != 0");

    if(mysql_num_rows(os_akill_res))
    {
	AKill *akill;
	gchar *user, *host;

	while((os_akill_row = mysql_fetch_row(os_akill_res)))
	{
	    akill = g_new0(AKill, 1);

	    user = os_akill_row[1];
	    host = strchr(user, '@');

	    if(!host)
	    {
		g_critical_syslog("Wrong akill format in DB: '%s' ( missing '@' ). ID: %s",
			os_akill_row[1], os_akill_row[0]);
		return;
	    }
	    *host++ = '\0';

	    akill->id = atoi(os_akill_row[0]);
	    akill->user = g_strdup(user);
	    akill->host = g_strdup(host);
	    akill->expire_at = atoi(os_akill_row[2]);
	    akill->timeout = g_timeout_source_add((akill->expire_at - NOW) gs_seconds,
		    (GSourceFunc) os_expire_akill, akill);

	    active_akills = g_slist_prepend(active_akills, akill);
	}
    }

    mysql_free_result(os_akill_res);
}

static G_INLINE_FUNC void os_akill_add_internal(User *caller, gchar **parv, gint expiry)
{
    gchar sqlmask[SQLMASKLEN + 1];
    gchar *mask = parv[0], *user, *host;
    gchar *reason = parv[1];
    gint id;
 
    if(strchr(mask, '!'))
    {
	send_out(":%s NOTICE %s :Invalid character in mask ( ! )",
		os->name, caller->name);
	return;
    }
    
    host = strchr(user = mask, '@');
    if(!host)
    {
	send_out(":%s NOTICE %s :Wrong akill format ( missing @ )",
		os->name, caller->name);
	return;
    }
    match_to_sqlmask(mask, sqlmask, SQLMASKLEN);
    *host++ = '\0';

    if(!strchr(host, '.'))
    {
	send_out(":%s NOTICE %s :Missing dot ( . ) in hostname",
		os->name,caller->name);
	return;
    }

    sql_query_assert(&os_akill_res, "SELECT `id` FROM `akill` WHERE `mask` = '%s'",
	    sqlmask);

    if(mysql_num_rows(os_akill_res))
    {
	os_akill_row = mysql_fetch_row(os_akill_res);
	
	send_out(":%s NOTICE %s :Mask %s@%s is already present on akill list (ID: %s)",
		os->name, caller->name, user, host, os_akill_row[0]);

	mysql_free_result(os_akill_res);
	return;
    }

    mysql_free_result(os_akill_res);

    /* OK, insert into the table
     */
    sql_query_assert(NULL, "INSERT INTO `akill` VALUES( 0, '%s', '%s@%s', '%s', '%s', NOW(), FROM_UNIXTIME(%ld) )",
	    sqlmask, user, host, reason, caller->name, expiry ? (NOW + expiry) : 0);
    id = sql_insert_id();

    send_out("AKILL %s %s %d %s %ld :%s [AKill ID: %d]",
	    host, user, expiry, caller->name, NOW, reason, id);
    send_out(":%s NOTICE %s :%s@%s added to the AKILL List",
	    os->name, caller->name, user, host);

    if(expiry)
    {
	AKill *akill = g_new0(AKill, 1);
	gchar expiration_time_s[128];

	akill->user = g_strdup(user);
	akill->host = g_strdup(host);
	akill->id = id;
	akill->expire_at = NOW + expiry;

	pretty_expiration_time(expiry, expiration_time_s, sizeof(expiration_time_s));
	send_out(":%s GLOBOPS :%s (%s@%s) added an AKILL for %s@%s "
		"[Reason: '%s'] [Expiration: %s] (ID: %d)",
		os->name, caller->name, caller->username, caller->host,
		user, host, reason, expiration_time_s, id);

	akill->timeout = g_timeout_source_add(expiry gs_seconds, (GSourceFunc) os_expire_akill, akill);

	g_mutex_lock(os_akill_mutex);
	active_akills = g_slist_prepend(active_akills, akill);
	g_mutex_unlock(os_akill_mutex);
    }
    else
    {
	send_out(":%s GLOBOPS :%s (%s@%s) added a permanent AKILL for %s@%s "
		"[Reason: '%s'] (ID: %d)",
		os->name, caller->name, caller->username, caller->host,
		user, host, reason, id);
    }
}

static gint os_akill_parse_time(gchar *timestr)
{
    gchar *s, *lastnum;
    gint ret = 0;

    s = lastnum = timestr;
    while(*s)
    {
	if(g_ascii_isdigit(*s))
	{
	    s++;
	    continue;
	}

	if(g_ascii_isalpha(*s))
	{
	    gchar c = *s;
	    gint number_of;
	    
	    *s++ = '\0';
	    if(!(number_of = atoi(lastnum)))
	    {
		return -1;
	    }
	    lastnum = s;

	    switch(c)
	    {
		case 'w':
		    ret += (number_of weeks);
		    break;
		case 'd':
		    ret += (number_of days);
		    break;
		case 'h':
		    ret += (number_of hours);
		    break;
		case 'm':
		    ret += (number_of minutes);
		    break;
		default:
		    return -1;
	    }

	    continue;
	}
		    
	return -1;
    }

    return ret;
}

static gboolean os_expire_akill(AKill *a)
{
    g_mutex_lock(os_akill_mutex);
    
    sql_query_assert_ret(FALSE, NULL, "DELETE FROM `akill` WHERE `id` = %d",
	    a->id);
    send_out(":%s GLOBOPS :Removing akill for %s@%s (Time expired)",
	    os->name, a->user, a->host);
    send_out("RAKILL %s %s", a->host, a->user);

    g_source_del(a->timeout);
    active_akills = g_slist_remove(active_akills, a);

    g_free(a->user);
    g_free(a->host);
    g_free(a);

    g_mutex_unlock(os_akill_mutex);
    
    return FALSE;
}

static gint akill_find(AKill *a, gint id)
{
    return((a->id == id) ? 0 : -1);
}

static gchar *pretty_expiration_time(time_t tm, gchar *retbuf, gsize len)
{
    guint n_weeks, n_days, n_hours, n_minutes, n_seconds;

    if((n_weeks = tm / (1 week)))
	tm -= n_weeks week;

    if((n_days = tm / (1 day)))
	tm -= n_days day;

    if((n_hours = tm / (1 hour)))
	tm -= n_hours hour;

    if((n_minutes = tm / (1 minute)))
	tm -= n_minutes minute;

    n_seconds = tm;

    if(n_weeks)
    {
	sux_snprintf(retbuf, len, "%d weeks %d days %d hours %d minutes %d seconds",
		n_weeks, n_days, n_hours, n_minutes, n_seconds);
    }
    else if(n_days)
    {
	sux_snprintf(retbuf, len, "%d days %d hours %d minutes %d seconds",
		n_days, n_hours, n_minutes, n_seconds);
    }
    else if(n_hours)
    {
	sux_snprintf(retbuf, len, "%d hours %d minutes %d seconds",
		n_hours, n_minutes, n_seconds);
    }
    else if(n_minutes)
    {
	sux_snprintf(retbuf, len, "%d minutes %d seconds",
		n_minutes, n_seconds);
    }
    else
    {
	sux_snprintf(retbuf, len, "%d seconds",
		n_seconds);
    }

    return retbuf;
}
