/*
 * Copyright 2002 Barnaba Marcello <vjt@azzurra.org>
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2a. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in
 *     the documentation and/or other materials provided with the
 *     distribution.
 * 2b. Redistribution in binary form requires specific prior written
 *     authorization of the maintainer.
 * 
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *    This product includes software developed by Barnaba Marcello.
 * 
 * 4. The names of the maintainer, developers and contributors may not be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE MAINTAINER, DEVELOPERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE DEVELOPERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: sql.c,v 1.30 2003/03/29 02:25:04 vjt Exp $ */

#include "sux.h"
#include "sql.h"
#include "log.h"
#include "conf.h"
#include "main.h"
#include "threads.h"
#include "sql-printf.h"
#include <mysql/errmsg.h>

static gboolean sql_try_reconnect(void);
static gboolean sql_try_recover(void);
static MYSQL *sql;
static GMutex *sql_mutex;

gboolean sql_connect(void)
{
    GNode *sec;
    gchar *host, *user, *pass, *port, *db;

    sec = find_section_id(SYMBOL_SECTION_SQL);
    if(!sec)
    {
	g_message("SQL Configuration section not found !!");
	return FALSE;
    }

    host = find_conf_id(sec, SYMBOL_SQL_HOST);
    user = find_conf_id(sec, SYMBOL_SQL_USER);
    pass = find_conf_id(sec, SYMBOL_SQL_PASS);
    port = find_conf_id(sec, SYMBOL_SQL_PORT);
    db = find_conf_id(sec, SYMBOL_SQL_DB);

    if(!host || !user || !pass || !port || !db)
    {
	g_message_syslog("Incomplete configuration !!");
	return FALSE;
    }

    sql = mysql_init(NULL);
    if(!sql)
    {
	g_message_syslog("Failed to initialize mysqlclient library");
	return FALSE;
    }

    mysql_real_connect(sql, host, user, pass, db, atoi(port), NULL,
	    CLIENT_IGNORE_SPACE);
    if(mysql_errno(sql))
    {
	g_message_syslog("Failed to connect to MySQL host: %s",
		mysql_error(sql));
	
	mysql_close(sql);
	return FALSE;
    }

    sql_mutex = g_mutex_new();

    /* okay, connected
     */
    g_message_syslog("SQL: Connected to %s [TID: %lu]",
	    mysql_get_host_info(sql), mysql_thread_id(sql));
    g_message_syslog("SQL: Protocol v.%d on server version %s",
	    mysql_get_proto_info(sql), mysql_get_server_info(sql));

    return TRUE;
}

void sql_shutdown(void)
{
    mysql_kill(sql, mysql_thread_id(sql));
    mysql_close(sql);
}

void sql_cleanup_before_exit(void)
{
    my_thread_end();
}

void sql_thread_init(void)
{
    my_thread_init();

    if(!mysql_thread_safe())
    {
	g_critical_syslog("The mysqlclient library is not compiled thread-safe !");
    }
}

G_INLINE_FUNC glong sql_affected_rows(void)
{
    return (glong) mysql_affected_rows(sql);
}

G_INLINE_FUNC guint sql_insert_id(void)
{
    return (guint) mysql_insert_id(sql);
}

G_INLINE_FUNC gchar *sql_get_server_info(void)
{
    return mysql_get_server_info(sql);
}

G_INLINE_FUNC glong sql_thread_id(void)
{
    return mysql_thread_id(sql);
}

G_INLINE_FUNC guint sql_get_proto_info(void)
{
    return mysql_get_proto_info(sql);
}

G_INLINE_FUNC gchar *sql_get_host_info(void)
{
    return mysql_get_host_info(sql);
}

static gboolean sql_try_reconnect(void)
{
    /* Connection lost ..
     */
    g_warning("Connection to mysql server %s lost (%s). Trying to reconnect ...",
	    mysql_get_host_info(sql), mysql_error(sql));
    mysql_kill(sql, mysql_thread_id(sql));

    mysql_close(sql);

    if(!sql_connect())
    {
	g_critical("Failed to reconnect to mysql server !");
	return FALSE;
    }

    g_message("Re-connection to %s established (TID: %lu, proto v.%d, server version %s)",
	    mysql_get_host_info(sql), mysql_thread_id(sql),
	    mysql_get_proto_info(sql), mysql_get_server_info(sql));

    return TRUE;
}

static gboolean sql_try_recover(void)
{
    switch(mysql_errno(sql))
    {
	case 0:
	    return TRUE;

	case CR_SERVER_GONE_ERROR:
	case CR_SERVER_LOST:
	    return sql_try_reconnect();
	default:
	    g_critical("Query failed: %s",
		    mysql_error(sql));
	    break;
    }

    return FALSE;
}

G_INLINE_FUNC gboolean __sql_query_assert(MYSQL_RES **rs, const gchar *q, ...)
{
    gushort i = 0, len;
    gchar qbuf[2048];
    va_list ap;
    gboolean ret = TRUE;
    
    va_start(ap, q);
    len = sql_vsprintf(qbuf, q, ap);
    va_end(ap);

    g_mutex_lock(sql_mutex);

    while(i++ < 5)
    {
	if(G_LIKELY((!mysql_real_query(sql, qbuf, len)) &&
		(rs == NULL || (*rs = mysql_store_result(sql)))))
	{
	    break;
	}

	g_warning("Query '%s': %s. Trying to recover ...",
		qbuf, mysql_error(sql) ? : "Unexpected NULL recordset");
	
	if(sql_try_recover())
	{
	    continue;
	}

	ret = FALSE;
	break;
    }

    g_mutex_unlock(sql_mutex);
    return ret;
}

G_INLINE_FUNC void sql_lock_tables(void)
{
    sql_query_assert(NULL, "LOCK TABLES `acl` WRITE, `akick` WRITE, `aop` WRITE, "
	    "`cfounder` WRITE, `chans` WRITE, `masks` WRITE, `memo` WRITE, "
	    "`nicks` WRITE, `sop` WRITE, `vop` WRITE, `nicks` AS `memo_rcpt` WRITE, "
	    "`akill` WRITE"); 
}

G_INLINE_FUNC void sql_unlock_tables(void)
{
    sql_query_assert(NULL, "UNLOCK TABLES");
}
